import math
import numpy
import matplotlib
import matplotlib.pyplot

import shared


########################################################################################################################
def Plot_Family_Tree(node_list, edge_list, plot_file_path):
    figure_size_unit = 1.0
    font_size_unit   = 1.0
    pad_size_unit    = 1.0

    figure_width  = shared.n_indiv_max * figure_size_unit + 6
    figure_height = shared.n_generation * figure_size_unit * 1.5 + 3
    figure = matplotlib.pyplot.figure(num=None, figsize=(figure_width,figure_height), dpi=300, facecolor='w')

    hm = 0.200 * 8 / figure_width
    w  = 1.000 - 0.250 * 8 / figure_width
    vm = 0.200 * 6 / figure_height
    h  = 1.000 - 0.600 * 6 / figure_height
    subplot = {}
    subplot["overall"]  = figure.add_subplot(1,1,1, position=[hm, vm, w, h], frame_on=True, zorder=110)
    subplot_id = "overall"


    fancybox = matplotlib.patches.FancyBboxPatch([0,0], 1, 1, transform=subplot[subplot_id].transAxes,
                                                 boxstyle='round,pad=0.0', linestyle='solid', linewidth=2.0,
                                                 edgecolor='black', facecolor='none', zorder=321)
    subplot[subplot_id].add_patch(fancybox)


    frame = subplot[subplot_id].patch
    frame.set_linewidth(2.00)
    frame.set_edgecolor('black')

    major_tick_width  =  2.00
    minor_tick_width  =  1.00
    major_tick_length = 10.00
    minor_tick_length =  5.00
    major_grid_width  =  2.00
    minor_grid_width  =  1.00
    major_grid_color  = '#E0E0E0'
    minor_grid_color  = '#E0E0E0'

    ###-----------------------------------------------------------------------------------------------------------------
    tickstep   = 1
    nminortick = 0

    x_min_view = 0
    x_max_view = shared.n_indiv_max + 1

    major_tick_locations = numpy.arange(x_min_view, x_max_view, dtype=int)
    minor_tick_locations = numpy.array([])

    major_tick_labels = ["%i" % x for x in major_tick_locations]
    major_tick_labels[0]  = ""
    axis_label = "Structure (in order of increasing energy)"

    subplot[subplot_id].xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(minor_tick_locations))
    subplot[subplot_id].xaxis.set_major_locator(matplotlib.ticker.FixedLocator(major_tick_locations))
    subplot[subplot_id].xaxis.set_ticklabels(major_tick_labels, minor=False)
    subplot[subplot_id].xaxis.set_tick_params(which='minor', direction='in', bottom=True, top=True, left=True, right=True, length=minor_tick_length, width=minor_tick_width, zorder=101)
    subplot[subplot_id].xaxis.set_tick_params(which='major', direction='in', bottom=True, top=True, left=True, right=True, length=major_tick_length, width=major_tick_width, pad=7*pad_size_unit, labelsize=24*font_size_unit, zorder=102)
    subplot[subplot_id].xaxis.set_label_text(axis_label, fontsize=28*font_size_unit, verticalalignment='top')
    subplot[subplot_id].xaxis.labelpad = 7 * pad_size_unit
    subplot[subplot_id].xaxis.grid(which='minor', color=minor_grid_color, linestyle='-', linewidth=minor_grid_width, zorder=101)
    subplot[subplot_id].xaxis.grid(which='major', color=major_grid_color, linestyle='-', linewidth=major_grid_width, zorder=102)
    subplot[subplot_id].set_xlim(left=x_min_view, right=x_max_view)

    ###-----------------------------------------------------------------------------------------------------------------
    tickstep   = 1
    nminortick = 0

    y_min_view = 0.0
    y_max_view = shared.n_generation + 1

    major_tick_locations = numpy.arange(y_min_view, y_max_view, dtype=int)
    minor_tick_locations = numpy.array([])

    major_tick_labels = ["%i" % x for x in numpy.arange(y_max_view, y_min_view, -1)]
    major_tick_labels[0]  = ""
    axis_label = "Generation"

    subplot[subplot_id].yaxis.set_minor_locator(matplotlib.ticker.FixedLocator(minor_tick_locations))
    subplot[subplot_id].yaxis.set_major_locator(matplotlib.ticker.FixedLocator(major_tick_locations))
    subplot[subplot_id].yaxis.set_ticklabels(major_tick_labels, minor=False)
    subplot[subplot_id].yaxis.set_tick_params(which='minor', direction='in', bottom=True, top=True, left=True, right=True, length=minor_tick_length, width=minor_tick_width, zorder=101)
    subplot[subplot_id].yaxis.set_tick_params(which='major', direction='in', bottom=True, top=True, left=True, right=True, length=major_tick_length, width=major_tick_width, pad=5*pad_size_unit, labelsize=24*font_size_unit, zorder=102)
    subplot[subplot_id].yaxis.set_label_text(axis_label, fontsize=28*font_size_unit, verticalalignment='center')
    subplot[subplot_id].yaxis.labelpad = 28 * pad_size_unit
    subplot[subplot_id].yaxis.grid(which='minor', color=minor_grid_color, linestyle='-', linewidth=minor_grid_width, zorder=101)
    subplot[subplot_id].yaxis.grid(which='major', color=major_grid_color, linestyle='-', linewidth=major_grid_width, zorder=102)
    subplot[subplot_id].set_ylim(bottom=y_min_view, top=y_max_view)

    ###-----------------------------------------------------------------------------------------------------------------
    origin_marker_dict = {"random":'^', "crossing":'s', "slice-mutation":'o', "atom-type-mutation":'p', "atom-position-mutation":'h', "survival":'D'}
    status_color_dict = {"surviving":'green', "live":'blue', "dead, unique":'red', "dead, duplicate":'gray'}
    origin_dash_dict = {"crossing":[8,0.1,8,0.1], "slice-mutation":[10,2,10,2], "atom-type-mutation":[2,2,2,2], "atom-position-mutation":[5,2,5,2], "survival":[10,2,2,2]}

    label1 = ["survival", "crossing", "slice-mutation", "atom-type-mutation", "atom-position-mutation", "random"]
    plot1  = []
    for operation in label1:
        plot1.append(subplot[subplot_id].plot([], [], linestyle="", marker=origin_marker_dict[operation], markersize=10.0, markeredgewidth=2.0, markeredgecolor='black', markerfacecolor='none', alpha=1.0, label=operation, zorder=122)[0])
    #

    label2 = ["surviving", "live", "dead, unique", "dead, duplicate"]
    plot2 = []
    for status in label2:
        plot2.append(subplot[subplot_id].plot([], [], linestyle="", marker='o', markersize=15.0, markeredgewidth=0.0, markeredgecolor='none', markerfacecolor=status_color_dict[status], alpha=1.0, label=status, zorder=122)[0])
    #

    label3 = label1[:-1]
    plot3 = []
    for operation in label3:
        plot3.append(subplot[subplot_id].plot([], [], linestyle="-", dashes=origin_dash_dict[operation], color='black', linewidth=1, marker='', alpha=1.0, label=operation, zorder=122)[0])
    #

    l1 = matplotlib.pyplot.legend(handles=plot1, title="Marker code:", loc='lower left', bbox_to_anchor=(0.00, shared.n_generation+1.1), bbox_transform=subplot[subplot_id].transData, ncol=1, columnspacing=3.0, labelspacing=0.40, borderaxespad=0.0, borderpad=0.5, numpoints=1, handlelength=1.0, handletextpad=0.5, fancybox=True, shadow=True)
    l1_frame = l1.get_frame()
    l1_frame.set_linewidth(2)
    l1_frame.set_facecolor('0.90')
    matplotlib.pyplot.gca().add_artist(l1)

    l2 = matplotlib.pyplot.legend(handles=plot2, title="Color code:", loc='lower left', bbox_to_anchor=((shared.n_indiv_max+1)/3 , shared.n_generation+1.1), bbox_transform=subplot[subplot_id].transData, ncol=1, columnspacing=3.0, labelspacing=0.40, borderaxespad=0.0, borderpad=0.5, numpoints=1, handlelength=1.0, handletextpad=0.5, fancybox=True, shadow=True)
    l2_frame = l2.get_frame()
    l2_frame.set_linewidth(2)
    l2_frame.set_facecolor('0.90')
    matplotlib.pyplot.gca().add_artist(l2)

    l3 = matplotlib.pyplot.legend(handles=plot3, title="Line code:", loc='lower left', bbox_to_anchor=((shared.n_indiv_max+1)/3*2 , shared.n_generation+1.1), bbox_transform=subplot[subplot_id].transData, ncol=1, columnspacing=3.0, labelspacing=0.40, borderaxespad=0.0, borderpad=0.5, numpoints=1, handlelength=3.0, handletextpad=0.5, fancybox=True, shadow=True)
    l3_frame = l3.get_frame()
    l3_frame.set_linewidth(2)
    l3_frame.set_facecolor('0.90')
    matplotlib.pyplot.gca().add_artist(l3)

    color_dict = {(True, "surviving"):'green', (True, "live"):'blue', (True, "dead"):'red', (False, "dead"):'gray'}
    for node in node_list:
        marker = origin_marker_dict[node[2]]
        color = color_dict[node[3]]
        x = node[1] + 1
        y = shared.n_generation - node[0]
        subplot[subplot_id].plot(x, y, linestyle="", marker=marker, markersize=10.0, markeredgewidth=2.0, markeredgecolor=color, markerfacecolor='none', alpha=1.0, zorder=122)
    #

    delta_x = 1.0 / shared.n_indiv_max
    margin_y = 0.2
    delta_y = (1.0 - 2 * margin_y) / shared.n_indiv_max

    i_color = -1
    color_list = [(x/5.0,0,0) for x in range(1, 6)]
    n_color = len(color_list)
    origin_dash_dict = {"crossing":[8,0.01,8,0.01], "slice-mutation":[8,2,8,2], "atom-type-mutation":[1,1,1,1], "atom-position-mutation":[3,2,3,2], "survival":[8,2,2,2]}
    n_edge = numpy.zeros((shared.n_generation,shared.n_indiv_max))
    deltay = 0.8
    i_generation0 = -1
    for edge in edge_list:
        i_generation = edge[1][0]
        i_indiv      = edge[1][1]

        if (i_generation != i_generation0):
            x_list = [1] * shared.n_indiv_max
            y_val = 0
            i_generation0 = i_generation
        #

        x0 = i_indiv
        x1 = x0

        y0 = shared.n_generation - i_generation
        y1 = y0 + margin_y + y_val * delta_y
        y_val = y_val + 1
        y2 = y1

        if (len(edge[2:]) > 1):
            dx = delta_x / 6
            dy = delta_y / 8
        else:
            dx = 0.0
            dy = 0.0
        #
        idx = -1
        idy = -1
        for (i_generation, i_indiv) in edge[2:]:
            if (x_list[i_indiv] % 2 == 0):
                x2 = i_indiv + x_list[i_indiv] / 2 * delta_x
            else:
                x2 = i_indiv - (x_list[i_indiv] + 1) / 2 * delta_x
            #
            x_list[i_indiv] = x_list[i_indiv] + 1

            x3 = x2
            y3 = shared.n_generation - i_generation

            xx = [x+1 for x in [x0+idx*dx, x1+idx*dx, x2, x3]]
            yy = [y0, y1+idy*dy, y2+idy*dy, y3]
            idx = idx + 2
            idy = idy + 2

            subplot[subplot_id].plot(xx, yy, linestyle='-', dashes=origin_dash_dict[edge[0]], marker='', linewidth=1.0, color='black', zorder=122)
            n_edge[i_generation,i_indiv] = n_edge[i_generation,i_indiv] + 1
        #
    #

    matplotlib.pyplot.savefig(plot_file_path)
#



