import numpy
import math
import os


########################################################################################################################
def Rotation_Matrix(i_axis, i_quarter):
    s = math.sin(i_quarter * math.pi / 2)
    c = math.cos(i_quarter * math.pi / 2)

    if (i_axis == 0):
        mat = numpy.array([[ 1,  0,  0],
                           [ 0,  c, -s],
                           [ 0,  s,  c]])
    elif (i_axis == 1):
        mat = numpy.array([[ c,  0,  s],
                           [ 0,  1,  0],
                           [-s,  0,  c]])
    elif (i_axis == 2):
        mat = numpy.array([[ c, -s,  0],
                           [ s,  c,  0],
                           [ 0,  0,  1]])
    #
    return mat
#

########################################################################################################################
def List_Directories(path=None):
    if (path is None):
        path = os.getcwd()
    #

    tmp_list = []
    if (os.path.isdir(path)):
        for entry in os.listdir(path):
            if (os.path.isdir(os.path.join(path, entry))):
                tmp_list.append(entry)
    #

    return sorted(tmp_list)
#

########################################################################################################################
def Linear_Tick_Calculator(minview, maxview, tickorigin=None, tickstep=None, nminortick=0, begin=None, end=None):
    if ((maxview - minview) / (abs(maxview) + abs(minview) + 1.0e-15) < 1.0e-12):
        print(":ERROR:   !!!   Too small view window.")
        exit(1)
    #
    if (tickorigin is None):
        tickorigin = minview
    #
    if (tickstep is None):
        major_tickstep = (maxview - minview) / 5
    else:
        major_tickstep = tickstep
    #
    if (major_tickstep == 0.0):
        print(":ERROR:   !!!   'tickstep == 0.0'.")
        exit(2)
    #
    if (major_tickstep / (maxview - minview) < 1.0e-3):
        print(":ERROR:   !!!   The given value for 'tickstep' is too small.")
        exit(3)
    #

    if (nminortick == 0):
        major_x1 = tickorigin + math.floor((minview - tickorigin) / major_tickstep) * major_tickstep
        minor_x1 = major_x1
    else:
        minor_tickstep = major_tickstep / (nminortick + 1)
        minor_x1 = tickorigin + math.floor((minview - tickorigin) / minor_tickstep) * minor_tickstep
        major_x1 = tickorigin + math.ceil((minor_x1 - tickorigin) / major_tickstep) * major_tickstep
    #

    if (nminortick == 0):
        major_x2 = tickorigin + math.ceil((maxview - tickorigin) / major_tickstep) * major_tickstep
        minor_x2 = major_x2
    else:
        minor_tickstep = major_tickstep / (nminortick + 1)
        minor_x2 = tickorigin + math.ceil((maxview - tickorigin) / minor_tickstep) * minor_tickstep
        major_x2 = tickorigin + math.floor((minor_x2 - tickorigin) / major_tickstep) * major_tickstep
    #

    if (begin == "major"):
        if (minor_x1 < major_x1):
            major_x1 = major_x1 - major_tickstep
        #
    elif (begin == "minor"):
        if (minor_x1 == major_x1):
            minor_x1 = minor_x1 - minor_tickstep
        #
    elif (begin is not None):
        print(":ERROR:   !!!   The given value for 'begin' is not defined.")
        exit(4)
    #

    if (end == "major"):
        if (minor_x2 > major_x2):
            major_x2 = major_x2 + major_tickstep
        #
    elif (end == "minor"):
        if (minor_x2 == major_x2):
            minor_x2 = minor_x2 + minor_tickstep
        #
    elif (end is not None):
        print(":ERROR:   !!!   The given value for 'end' is not defined.")
        exit(5)
    #


    major_tick_locations = numpy.arange(major_x1, major_x2+major_tickstep/2, major_tickstep)
    if (nminortick > 0):
        minor_tickstep = major_tickstep / (nminortick + 1)
        minor_tick_locations = numpy.arange(minor_x1, minor_x2+minor_tickstep/2, minor_tickstep)
    else:
        minor_tick_locations = numpy.array([])
    #

    return (major_tick_locations, minor_tick_locations)
#

########################################################################################################################





