import numpy
import math
import shared
import lj_parameters


###=====================================================================================================================
def Lennard_Jones_Potential(atomid, rpos):
    e = 0.0
    for i_atom in range(1, len(atomid)):
        for j_atom in range(0, i_atom):
            eps = lj_parameters.lj_e[atomid[i_atom],atomid[j_atom]]

            r0  = lj_parameters.lj_r[atomid[i_atom],atomid[j_atom]]
            rij = numpy.linalg.norm(rpos[i_atom,:] - rpos[j_atom,:]) * shared.a
            #print(rij)

            rm06 = (r0 / rij) ** 6
            e = e + eps * (rm06 ** 2 - 2 * rm06)
        #
    #
    return e
#

###=====================================================================================================================
def Lennard_Jones_Force(atomid, rpos):
    f = numpy.zeros(numpy.shape(rpos))
    for i_atom in range(1, len(atomid)):
        for j_atom in range(0, i_atom):
            eps = lj_parameters.lj_e[atomid[i_atom],atomid[j_atom]]

            r0  = lj_parameters.lj_r[atomid[i_atom],atomid[j_atom]]
            rij_vec = (rpos[i_atom,:] - rpos[j_atom,:]) * shared.a
            rij = numpy.linalg.norm(rij_vec)

            rm06 = (r0 / rij) ** 6
            fi = -12 * eps * (rm06 ** 2 - rm06) / (rij ** 2) * rij_vec
            f[i_atom,:] = f[i_atom,:] + fi
            f[j_atom,:] = f[j_atom,:] - fi
        #
    #
    return f
#

###=====================================================================================================================
def Structure_Energy(struct):
    return Lennard_Jones_Potential(shared.atom_id, struct.rpos)
#

###=====================================================================================================================
def Structure_Force(struct):
    return Lennard_Jones_Force(shared.atom_id, struct.rpos)
#

###=====================================================================================================================
def Structure_E(rpos):
    return Lennard_Jones_Potential(shared.atom_id, numpy.reshape(rpos, (shared.n_atom, 3)))
#

###=====================================================================================================================
def Structure_Der(rpos):
    return Lennard_Jones_Force(shared.atom_id, numpy.reshape(rpos, (shared.n_atom, 3))).reshape(-1)
#

###=====================================================================================================================


