###!/usr/bin/env python3
import math
import scipy
import scipy.optimize
import numpy
import copy
import pickle
import os

import shared
import initialize
import potentials
import save
import selection
import family_tree
import lj_parameters
import fitness
import fingerprint


random_state_file = "numpy_random_state.pickle"

if (not shared.restart_random_generator):
    if (os.path.exists(random_state_file)):
        with open(random_state_file, 'rb') as inf:
            numpy.random.set_state(pickle.load(inf))
        #
    else:
        print(":ERROR:   !!!   The specified file for random-generator state was not found.")
        exit()
    #
#

if (os.path.exists(random_state_file)):
    with open(random_state_file+".tmp", 'wb') as outf:
        pickle.dump(numpy.random.get_state(), outf)
    #
else:
    with open(random_state_file, 'wb') as outf:
        pickle.dump(numpy.random.get_state(), outf)
    #
#

lj_parameters.Set_Lennard_Jones_Parameters()


###---------------------------------------------------------------------------------------------------------------------
shared.i_generation = 0
shared.initial_indiv = initialize.Initialize()

for i_generation in range(0, shared.n_generation):
    shared.n_try_fill                   = 0
    shared.n_try_crossing               = 0
    shared.n_try_slice_mutation         = 0
    shared.n_try_atom_type_mutation     = 0
    shared.n_try_atom_position_mutation = 0
    shared.n_try_random                 = 0

    shared.i_generation = i_generation
    fitness.Evaluate(shared.initial_indiv)
    save.Save_Structures(shared.initial_indiv, "initial")

    shared.final_indiv = copy.deepcopy(shared.initial_indiv) 
    fitness.Relax(shared.final_indiv)
    fitness.Analyse(shared.final_indiv)
    save.Save_Structures(shared.final_indiv, "final", bests=True)
    shared.gen_indiv.append(shared.final_indiv)

    print("i_generation = %4i, %4i," % (i_generation+1, shared.n_live_uniq), end="")

    shared.initial_indiv = selection.Create_Offsprings()

    print("   %8i %8i %8i %8i %8i %8i" % (shared.n_try_fill, shared.n_try_crossing, shared.n_try_slice_mutation, shared.n_try_atom_type_mutation, shared.n_try_atom_position_mutation, shared.n_try_random))
#


###---------------------------------------------------------------------------------------------------------------------
if (shared.n_indiv * shared.n_generation <= 1000):
    print()
    print()
    node_list = []
    edge_list = []
    for i_generation in list(reversed(range(0, shared.n_generation))):
        for i_indiv in range(0, len(shared.gen_indiv[i_generation])):
            node_list.append((i_generation, shared.gen_indiv[i_generation][i_indiv].i_indiv_energy, shared.gen_indiv[i_generation][i_indiv].origin, (shared.gen_indiv[i_generation][i_indiv].unique, shared.gen_indiv[i_generation][i_indiv].status), shared.gen_indiv[i_generation][i_indiv].energy))
            if (shared.gen_indiv[i_generation][i_indiv].back is not None):
                edge_list.append([shared.gen_indiv[i_generation][i_indiv].origin, (i_generation, shared.gen_indiv[i_generation][i_indiv].i_indiv_energy)])
                tmp_list = []
                for backid in shared.gen_indiv[i_generation][i_indiv].back:
                    tmp_list.append((i_generation-1, shared.gen_indiv[i_generation-1][backid].i_indiv_energy))
                #
                if (len(tmp_list) > 0):
                    edge_list[-1] = edge_list[-1] + sorted(tmp_list, key=lambda x: x[1])
                #
            #
        #
    #
    family_tree.Plot_Family_Tree(node_list, edge_list, "family_tree.pdf")
#

###---------------------------------------------------------------------------------------------------------------------


