import os
import numpy
import copy

import shared
import variation_operations
import structure
import save


###=====================================================================================================================
class Roulette_Wheel:
    ###=================================================================================================================
    def __init__(self, energy_array):
        self.n = len(energy_array)
        energy_min = numpy.min(energy_array)
        energy_max = numpy.max(energy_array)
        mating_temperature = energy_max - energy_min
        if (mating_temperature == 0.0):
            self.bound = numpy.linspace(0.0, 1.0, num=self.n, endpoint=True)
        else:
            self.weight = numpy.exp(-(energy_array - energy_min) / mating_temperature)
            self.weight = self.weight / numpy.sum(self.weight)
            self.bound = numpy.cumsum(self.weight)
        #
    #

    ###=================================================================================================================
    def Select(self, n_select):
        if (len(self.bound) == 1):
            return numpy.array([0], dtype=int)
        else:
            if (n_select > len(self.bound)):
                print(":ERROR:   !!!   'n_select > len(self.bound)'.")
                exit()
            else:
                r = numpy.random.random(n_select)
                i_set = set(numpy.searchsorted(self.bound, r, side='left', sorter=None))
                if (len(i_set) < n_select):
                    weight_array = numpy.copy(self.weight)
                    weight_array[list(i_set)] = 0.0
                    while (len(i_set) < n_select):
                        bound_array = numpy.cumsum(weight_array / numpy.sum(weight_array))
                        r = numpy.random.random(1)
                        i = numpy.searchsorted(bound_array, r, side='left', sorter=None)[0]
                        i_set.add(i)
                        weight_array[i] = 0.0
                    #
                #
            #
            return numpy.array(list(i_set))
        #
    #
#

###=====================================================================================================================
def Create_Offsprings():
    live_indx_array = shared.i_indiv_array
    selector = Roulette_Wheel(shared.energy_array)

    next_generation = shared.i_generation + 2
    i_generation_next = shared.i_generation + 1
    i_offspring = -1
    offspring = []

    ###-----------------------------------------------------------------------------------------------------------------
    for i_crossing in range(0, shared.n_crossing):
        i_offspring = i_offspring + 1
        shared.i_indiv_overall = shared.i_indiv_overall + 1
        while (True):
            i_list = selector.Select(2)
            (success, struct) = variation_operations.Crossing(i_generation_next, i_offspring, shared.i_indiv_overall, shared.final_indiv[live_indx_array[i_list[0]]], shared.final_indiv[live_indx_array[i_list[1]]])
            if (success):
                offspring.append(struct)
                break
            #
        #
    #

    ###-----------------------------------------------------------------------------------------------------------------
    for i_slice_mutation in range(0, shared.n_slice_mutation):
        i_offspring = i_offspring + 1
        shared.i_indiv_overall = shared.i_indiv_overall + 1
        while (True):
            i = selector.Select(1)[0]
            (success, struct) = variation_operations.Slice_Mutation(i_generation_next, i_offspring, shared.i_indiv_overall, shared.final_indiv[live_indx_array[i]])
            if (success):
                offspring.append(struct)
                break
            #
        #
    #

    ###-----------------------------------------------------------------------------------------------------------------
    for i_atom_type_mutation in range(0, shared.n_atom_type_mutation):
        i_offspring = i_offspring + 1
        shared.i_indiv_overall = shared.i_indiv_overall + 1
        while (True):
            i = selector.Select(1)[0]
            (success, struct) = variation_operations.Atom_Type_Mutation(i_generation_next, i_offspring, shared.i_indiv_overall, shared.final_indiv[live_indx_array[i]])
            if (success):
                offspring.append(struct)
                break
            #
        #
    #

    ###-----------------------------------------------------------------------------------------------------------------
    for i_atom_position_mutation in range(0, shared.n_atom_position_mutation):
        i_offspring = i_offspring + 1
        shared.i_indiv_overall = shared.i_indiv_overall + 1
        while (True):
            i = selector.Select(1)[0]
            (success, struct) = variation_operations.Atom_Position_Mutation(i_generation_next, i_offspring, shared.i_indiv_overall, shared.final_indiv[live_indx_array[i]])
            if (success):
                offspring.append(struct)
                break
            #
        #
    #

    ###-----------------------------------------------------------------------------------------------------------------
    for i_random in range(0, shared.n_random):
        i_offspring = i_offspring + 1
        shared.i_indiv_overall = shared.i_indiv_overall + 1
        struct = variation_operations.Random_Structure(i_generation_next, i_offspring, shared.i_indiv_overall)
        offspring.append(struct)
    #

    ###-----------------------------------------------------------------------------------------------------------------
    for i_surviving in range(0, shared.n_surviving):
        i_offspring = i_offspring + 1
        struct = variation_operations.Survival(i_generation_next, i_offspring, shared.final_indiv[live_indx_array[i_surviving]])
        offspring.append(struct)
    #

    return offspring
#

###=====================================================================================================================


