import numpy

import shared
import fingerprint
import atomic_data


class Structure:
    ####=================================================================================================================
    def __init__(self, title=None):
        if (title is None):
            self.title = "Untitled"
        else:
            self.title = str(title)
        #
        self.i_generation    = None
        self.i_indiv         = None
        self.i_indiv_overall = None
        self.i_indiv_energy  = None
        self.i_category      = None
        self.origin          = None
        self.history         = []
        self.back            = None

        self.rpos            = None
        self.fp_key          = None
        self.fp_val          = None

        self.relaxed         = None
        self.energy          = None
        self.force           = None
        self.unique          = None
        self.status          = None
    #

    ###=================================================================================================================
    def Set_Center(self):
        min_rpos = numpy.amin(self.rpos, axis=0)
        max_rpos = numpy.amax(self.rpos, axis=0)
        shift = 0.5 - (min_rpos + max_rpos) * 0.5
        self.rpos = self.rpos + shift
    #

    ###=================================================================================================================
    def Fill_Randomly(self):
        self.rpos = numpy.empty((shared.n_atom, 3))
        for i_atom in range(0, shared.n_atom):
            n_try = 0
            while (True):
                self.rpos[i_atom,:] = shared.margin +  numpy.random.random(3) * (1.0 - 2 * shared.margin)

                accept = True
                for j_atom in range(0, i_atom):
                    rij = numpy.linalg.norm(self.rpos[i_atom,:] - self.rpos[j_atom,:]) * shared.a
                    if (rij < shared.min_atomic_distance):
                        accept = False
                        break
                    #
                #
                if (accept):
                    break
                #
                n_try = n_try + 1
                if (n_try == shared.n_try_fill_max):
                    print(":ERROR:   !!!   n_try exceeded n_try_random_fill_max.")
                    raise
                    exit()
                #
            #
            shared.n_try_fill = shared.n_try_fill + n_try
        #
        self.Set_Center()
    #

    ###=================================================================================================================
    def Save_XSF(self, outputfile):
        self.Save_Periodic_XSF(outputfile)
    #

    ###=================================================================================================================
    def Save_Periodic_XSF(self, outputfile):
        with open(outputfile, 'w') as outf:
            print("CRYSTAL", file=outf)
            print("#", file=outf)
            print("PRIMVEC", file=outf)
            print("%20.12f %20.12f %20.12f" % (shared.a, 0.0, 0.0), file=outf)
            print("%20.12f %20.12f %20.12f" % (0.0, shared.a, 0.0), file=outf)
            print("%20.12f %20.12f %20.12f" % (0.0, 0.0, shared.a), file=outf)
            print("#", file=outf)
            print("PRIMCOORD", file=outf)
            print("%i   1" % shared.n_atom, file=outf)
            for i_atom in range(0, shared.n_atom):
                print("%3i   %20.12f %20.12f %20.12f   %20.12f %20.12f %20.12f" % ((shared.sorted_atom_numb[i_atom],)+tuple([x * shared.a for x in self.rpos[i_atom,:]])+tuple(self.force[i_atom,:])), file=outf)
            #
            print("#", file=outf)
            print("#", file=outf)
            print("#", file=outf)
            print("### title           = %s" % (self.title), file=outf)
            print("### i_generation    = %i" % (self.i_generation+1), file=outf)
            print("### i_indiv         = %i" % (self.i_indiv+1), file=outf)
            print("### i_indiv_overall = %i" % (self.i_indiv_overall+1), file=outf)
            print("### origin          = %s" % (str(self.origin)), file=outf)
            for i_history in range(0, len(self.history)):
                print("### history         : %s" % (self.history[i_history]), file=outf)
            #
            print("### back            = %s" % (str(self.back)), file=outf)
            print("#", file=outf)
            print("### sides           = %20.12f %20.12f %20.12f" % (shared.a, shared.a, shared.a), file=outf)
            print("### angles          = %20.12f %20.12f %20.12f" % (90.0, 90.0, 90.0), file=outf)
            print("### volume          = %20.12f" % (shared.a ** 3), file=outf)
            print("#", file=outf)
            print("### relaxed         = %s" % str(self.relaxed), file=outf)
            print("### energy          = %20.12f" % (self.energy), file=outf)
            print("### unique          = %s" % (str(self.unique)), file=outf)
            print("### status          = %s" % (str(self.status)), file=outf)
        #
    #

    ###=================================================================================================================
    def Save_Atomic_XSF(self, outputfile):
        with open(outputfile, 'w') as outf:
            print("### CRYSTAL", file=outf)
            print("####", file=outf)
            print("### PRIMVEC", file=outf)
            print("### %20.12f %20.12f %20.12f" % (shared.a, 0.0, 0.0), file=outf)
            print("### %20.12f %20.12f %20.12f" % (0.0, shared.a, 0.0), file=outf)
            print("### %20.12f %20.12f %20.12f" % (0.0, 0.0, shared.a), file=outf)
            print("####", file=outf)
            print("### PRIMCOORD", file=outf)
            print("### %i   1" % shared.n_atom, file=outf)
            print("ATOM", file=outf)
            for i_atom in range(0, shared.n_atom):
                print("%3i   %20.12f %20.12f %20.12f   %20.12f %20.12f %20.12f" % ((shared.sorted_atom_numb[i_atom],)+tuple([x * shared.a for x in self.rpos[i_atom,:]])+tuple(self.force[i_atom,:])), file=outf)
            #
            print("#", file=outf)
            print("#", file=outf)
            print("#", file=outf)
            print("### title           = %s" % (self.title), file=outf)
            print("### i_generation    = %i" % (self.i_generation+1), file=outf)
            print("### i_indiv         = %i" % (self.i_indiv+1), file=outf)
            print("### i_indiv_overall = %i" % (self.i_indiv_overall+1), file=outf)
            print("### origin          = %s" % (str(self.origin)), file=outf)
            for i_history in range(0, len(self.history)):
                print("### history         : %s" % (self.history[i_history]), file=outf)
            #
            print("### back            = %s" % (str(self.back)), file=outf)
            print("#", file=outf)
            print("### sides           = %20.12f %20.12f %20.12f" % (shared.a, shared.a, shared.a), file=outf)
            print("### angles          = %20.12f %20.12f %20.12f" % (90.0, 90.0, 90.0), file=outf)
            print("### volume          = %20.12f" % (shared.a ** 3), file=outf)
            print("#", file=outf)
            print("### relaxed         = %s" % str(self.relaxed), file=outf)
            print("### energy          = %20.12f" % (self.energy), file=outf)
            print("### unique          = %s" % (str(self.unique)), file=outf)
            print("### status          = %s" % (str(self.status)), file=outf)
        #
    #

    ####================================================================================================================
    def Calculate_RDF_Like_Fingerprint(self):
        (self.fp_key, self.fp_val) = fingerprint.RDF_Like_Fingerprint(shared.atom_id, self.rpos)
    #

    ####================================================================================================================
    def Save_Fingerprint(self, filename):
        (i1, i2) = self.fp_key[0]
        numb1 = shared.uniq_atom_numb[i1]
        numb2 = shared.uniq_atom_numb[i2]
        symb1 = atomic_data.atom_list[numb1].symb
        symb2 = atomic_data.atom_list[numb2].symb
        header = "# %4i %4i     %s(%i)   %s(%i)" % (i1, i2, symb1, numb1, symb2, numb2)
        data = numpy.vstack((shared.r_array, self.fp_val[0]))
        for i in range(1, len(self.fp_key)):
            (i1, i2) = self.fp_key[i]
            numb1 = shared.uniq_atom_numb[i1]
            numb2 = shared.uniq_atom_numb[i2]
            symb1 = atomic_data.atom_list[numb1].symb
            symb2 = atomic_data.atom_list[numb2].symb
            header = header + "\n# %4i %4i     %s(%i)   %s(%i)" % (i1, i2, symb1, numb1, symb2, numb2)
            data = numpy.vstack((data, self.fp_val[i]))
        #
        numpy.savetxt(filename, numpy.transpose(data), fmt='%24.15e', header=header)
    #

#########################################################################################################################


