import numpy
import copy

import shared
import structure
import fingerprint
import hard_constraints
import functions


###=====================================================================================================================
def Random_Structure(i_generation, i_indiv, i_indiv_overall):
    title = "%6i:   %4.4i.%4.4i" % (i_indiv_overall+1, i_generation+1, i_indiv+1)
    struct = structure.Structure(title=title)

    n_try = 0
    while (True):
        struct.Fill_Randomly()

        if (hard_constraints.Check_Bond_Length(shared.atom_id, struct.rpos)):
            (fp_key, fp_val) = fingerprint.RDF_Like_Fingerprint(shared.atom_id, struct.rpos)
            if (fingerprint.Is_Unique(fp_val)):
                shared.fingerprint_archive.append(fp_val)
                success = True
                break
            #
        #
        n_try = n_try + 1
        if (n_try == shared.n_try_random_max):
            success = False
            break
        #
    #
    shared.n_try_random = shared.n_try_random + n_try

    if (success):
        struct.fp_key = numpy.copy(fp_key)
        struct.fp_val = numpy.copy(fp_val)

        struct.i_generation    = i_generation
        struct.i_indiv         = i_indiv
        struct.i_indiv_overall = i_indiv_overall

        history = "%4.4i.%4.4i   Random ()" % (i_generation+1, i_indiv+1)
        struct.history.append(history)
        struct.origin = "random"
        struct.back   = None

        return struct
    else:
        print(":ERROR:   !!!   Failed to create a unique random structure.")
        exit()
    #
#

###=====================================================================================================================
def Survival(i_generation, i_indiv, struct0):
    struct = copy.deepcopy(struct0)

    title = "%6i:   %4.4i.%4.4i" % (struct0.i_indiv_overall+1, i_generation+1, i_indiv+1)
    struct.title = title

    struct.i_generation    = i_generation
    struct.i_indiv         = i_indiv
    struct.i_indiv_energy  = None

    struct.fp_dist = None
    struct.status  = None

    history = "%4.4i.%4.4i   Survival (%4.4i)" % (i_generation+1, i_indiv+1, struct0.i_indiv+1)
    struct.history.append(history)
    struct.origin = "survival"
    struct.back   = [struct0.i_indiv]

    return struct
#

###=====================================================================================================================
def Crossing(i_generation, i_indiv, i_indiv_overall, struct0, struct1):
    n_try = 0
    while (True):
        tmp_list = []
        i_axis    = numpy.random.randint(low=0, high=3)
        i_quarter = numpy.random.randint(low=0, high=4)
        rpos0 = numpy.matmul(functions.Rotation_Matrix(i_axis, i_quarter), struct0.rpos.transpose()).transpose()
        tmp_list = tmp_list + [i_axis+1, i_quarter+1]

        i_axis    = numpy.random.randint(low=0, high=3)
        i_quarter = numpy.random.randint(low=0, high=4)
        rpos1 = numpy.matmul(functions.Rotation_Matrix(i_axis, i_quarter), struct1.rpos.transpose()).transpose()
        tmp_list = tmp_list + [i_axis+1, i_quarter+1]

        i_axis = numpy.random.randint(low=0, high=3)
        x_min = min(numpy.amin(rpos0[:,i_axis]), numpy.amin(rpos1[:,i_axis]))
        x_max = max(numpy.amax(rpos0[:,i_axis]), numpy.amax(rpos1[:,i_axis]))
        delta = x_max - x_min
        x = numpy.random.uniform(low=x_min+0.25*delta, high=x_max-0.25*delta)
        tmp_list = tmp_list + [i_axis+1, x]

        rpos = numpy.empty((shared.n_atom, 3))
        i_rpos = -1
        for atom_id in range(0, shared.n_uniq_atom):
            indx = [None, None]
            indx[0] = [(0,i_atom) for i_atom in numpy.nonzero((rpos0[:,i_axis] < x)  * (shared.atom_id == atom_id))[0]]
            indx[1] = [(1,i_atom) for i_atom in numpy.nonzero((rpos1[:,i_axis] >= x) * (shared.atom_id == atom_id))[0]]
            selected_atom_list = indx[0] + indx[1]

            while (len(selected_atom_list) > shared.uniq_atom_count[atom_id]):
                i_atom = numpy.random.randint(low=0, high=len(selected_atom_list))
                del selected_atom_list[i_atom]
            #

            if (len(selected_atom_list) < shared.uniq_atom_count[atom_id]):
                indx[0] = [(0,i_atom) for i_atom in numpy.nonzero((rpos0[:,i_axis] >= x)  * (shared.atom_id == atom_id))[0]]
                indx[1] = [(1,i_atom) for i_atom in numpy.nonzero((rpos1[:,i_axis] <  x)  * (shared.atom_id == atom_id))[0]]
                remaining_atom_list = indx[0] + indx[1]
                while (len(selected_atom_list) < shared.uniq_atom_count[atom_id]):
                    i_atom = numpy.random.randint(low=0, high=len(remaining_atom_list))
                    selected_atom_list.append(remaining_atom_list[i_atom])
                    del remaining_atom_list[i_atom]
                #
            #
            for item in selected_atom_list:
                i_rpos = i_rpos + 1
                if (item[0] == 0):
                    rpos[i_rpos,:] = rpos0[item[1],:]
                elif (item[0] == 1):
                    rpos[i_rpos,:] = rpos1[item[1],:]
                #
            #
        #

        if (hard_constraints.Check_Bond_Length(shared.atom_id, rpos)):
            (fp_key, fp_val) = fingerprint.RDF_Like_Fingerprint(shared.atom_id, rpos)
            if (fingerprint.Is_Unique(fp_val)):
                shared.fingerprint_archive.append(fp_val)
                success = True
                break
            #
        #
        n_try = n_try + 1
        if (n_try == shared.n_try_crossing_max):
            success = False
            break
        #
    #
    shared.n_try_crossing = shared.n_try_crossing + n_try

    if (success):
        title = "%6i:   %4.4i.%4.4i" % (i_indiv_overall+1, i_generation+1, i_indiv+1)
        struct = structure.Structure(title=title)

        struct.rpos   = numpy.copy(rpos)
        struct.fp_key = numpy.copy(fp_key)
        struct.fp_val = numpy.copy(fp_val)

        struct.i_generation    = i_generation
        struct.i_indiv         = i_indiv
        struct.i_indiv_overall = i_indiv_overall

        history = "%4.4i.%4.4i   Crossing   (%4.4i, %4.4i)   %i %i   %i %i   %i   %12.9f" % ((i_generation+1, i_indiv+1, struct0.i_indiv+1, struct1.i_indiv+1) + tuple(tmp_list))
        struct.history.append(history)
        struct.origin = "crossing"
        struct.back   = [struct0.i_indiv, struct1.i_indiv]

        struct.Set_Center()
    else:
        struct = None
    #

    return (success, struct)
#

###=====================================================================================================================
def Slice_Mutation(i_generation, i_indiv, i_indiv_overall, struct0):
    n_try = 0
    while (True):
        tmp_list = []
        i_axis    = numpy.random.randint(low=0, high=3)
        i_quarter = numpy.random.randint(low=0, high=4)
        rpos0 = numpy.matmul(functions.Rotation_Matrix(i_axis, i_quarter), struct0.rpos.transpose()).transpose()
        tmp_list = tmp_list + [i_axis+1, i_quarter+1]

        i_axis = numpy.random.randint(low=0, high=3)
        i_quarter = numpy.random.randint(low=0, high=4)
        rpos1 = numpy.matmul(functions.Rotation_Matrix(i_axis, i_quarter), struct0.rpos.transpose()).transpose()
        tmp_list = tmp_list + [i_axis+1, i_quarter+1]

        i_axis = numpy.random.randint(low=0, high=3)
        x_min = min(numpy.amin(rpos0[:,i_axis]), numpy.amin(rpos1[:,i_axis]))
        x_max = max(numpy.amax(rpos0[:,i_axis]), numpy.amax(rpos1[:,i_axis]))
        delta = x_max - x_min
        x = numpy.random.uniform(low=x_min+0.25*delta, high=x_max-0.25*delta)
        tmp_list = tmp_list + [i_axis+1, x]

        rpos = numpy.empty((shared.n_atom, 3))
        i_rpos = -1
        for atom_id in range(0, shared.n_uniq_atom):
            indx = [None, None]
            indx[0] = [(0,i_atom) for i_atom in numpy.nonzero((rpos0[:,i_axis] < x)  * (shared.atom_id == atom_id))[0]]
            indx[1] = [(1,i_atom) for i_atom in numpy.nonzero((rpos1[:,i_axis] >= x) * (shared.atom_id == atom_id))[0]]
            selected_atom_list = indx[0] + indx[1]

            while (len(selected_atom_list) > shared.uniq_atom_count[atom_id]):
                i_atom = numpy.random.randint(low=0, high=len(selected_atom_list))
                del selected_atom_list[i_atom]
            #

            if (len(selected_atom_list) < shared.uniq_atom_count[atom_id]):
                indx[0] = [(0,i_atom) for i_atom in numpy.nonzero((rpos0[:,i_axis] >= x)  * (shared.atom_id == atom_id))[0]]
                indx[1] = [(1,i_atom) for i_atom in numpy.nonzero((rpos1[:,i_axis] <  x)  * (shared.atom_id == atom_id))[0]]
                remaining_atom_list = indx[0] + indx[1]
                while (len(selected_atom_list) < shared.uniq_atom_count[atom_id]):
                    i_atom = numpy.random.randint(low=0, high=len(remaining_atom_list))
                    selected_atom_list.append(remaining_atom_list[i_atom])
                    del remaining_atom_list[i_atom]
                #
            #
            for item in selected_atom_list:
                i_rpos = i_rpos + 1
                if (item[0] == 0):
                    rpos[i_rpos,:] = rpos0[item[1],:]
                elif (item[0] == 1):
                    rpos[i_rpos,:] = rpos1[item[1],:]
                #
            #
        #

        if (hard_constraints.Check_Bond_Length(shared.atom_id, rpos)):
            (fp_key, fp_val) = fingerprint.RDF_Like_Fingerprint(shared.atom_id, rpos)
            if (fingerprint.Is_Unique(fp_val)):
                shared.fingerprint_archive.append(fp_val)
                success = True
                break
            #
        #
        n_try = n_try + 1
        if (n_try == shared.n_try_slice_mutation_max):
            success = False
            break
        #
    #
    shared.n_try_slice_mutation = shared.n_try_slice_mutation + n_try

    if (success):
        title = "%6i:   %4.4i.%4.4i" % (i_indiv_overall+1, i_generation+1, i_indiv+1)
        struct = structure.Structure(title=title)

        struct.rpos   = numpy.copy(rpos)
        struct.fp_key = numpy.copy(fp_key)
        struct.fp_val = numpy.copy(fp_val)

        struct.i_generation    = i_generation
        struct.i_indiv         = i_indiv
        struct.i_indiv_overall = i_indiv_overall

        history = "%4.4i.%4.4i   Slice_Mutation   (%4.4i)   %i %i   %i %i   %i   %12.9f" % ((i_generation+1, i_indiv+1, struct0.i_indiv+1) + tuple(tmp_list))
        struct.history.append(history)
        struct.origin = "slice-mutation"
        struct.back = [struct0.i_indiv]

        struct.Set_Center()
    else:
        struct = None
    #

    return (success, struct)
#

###=====================================================================================================================
def Atom_Type_Mutation(i_generation, i_indiv, i_indiv_overall, struct0):
    n_try = 0
    n_swap = 1 + numpy.random.randint(low=0, high=shared.n_swap_max)
    while (True):
        rpos = numpy.copy(struct0.rpos)
        swap_list = []
        for i_swap in range(0, n_swap):
            tmp_array = numpy.arange(0, shared.n_uniq_atom, dtype=int)
            numpy.random.shuffle(tmp_array)
            [i1, i2] = tmp_array[0:2]
            tmp_array = numpy.nonzero(shared.atom_id == i1)[0]
            numpy.random.shuffle(tmp_array)
            i1_atom = tmp_array[0]

            tmp_array = numpy.nonzero(shared.atom_id == i2)[0]
            numpy.random.shuffle(tmp_array)
            i2_atom = tmp_array[0]

            tmp_rpos = numpy.copy(rpos[i1_atom,:])
            rpos[i1_atom,:] = rpos[i2_atom,:]
            rpos[i2_atom,:] = tmp_rpos
            swap_list.append((i1_atom,i2_atom))
        #
        if (hard_constraints.Check_Bond_Length(shared.atom_id, rpos)):
            (fp_key, fp_val) = fingerprint.RDF_Like_Fingerprint(shared.atom_id, rpos)
            if (fingerprint.Is_Unique(fp_val)):
                shared.fingerprint_archive.append(fp_val)
                success = True
                break
            #
        #
        n_try = n_try + 1
        if (n_try == shared.n_try_atom_type_mutation_max):
            success = False
            break
        #
    #
    shared.n_try_atom_type_mutation = shared.n_try_atom_type_mutation + n_try

    if (success):
        title = "%6i:   %4.4i.%4.4i" % (i_indiv_overall+1, i_generation+1, i_indiv+1)
        struct = structure.Structure(title=title)

        struct.rpos   = numpy.copy(rpos)
        struct.fp_key = numpy.copy(fp_key)
        struct.fp_val = numpy.copy(fp_val)

        struct.i_generation    = i_generation
        struct.i_indiv         = i_indiv
        struct.i_indiv_overall = i_indiv_overall

        history = "%4.4i.%4.4i   Atom_Type_Mutation   (%i)   %s" % (i_generation+1, i_indiv+1, struct0.i_indiv+1, str(swap_list))
        struct.history.append(history)
        struct.origin = "atom-type-mutation"
        struct.back = [struct0.i_indiv]

        struct.Set_Center()
    else:
        struct = None
    #

    return (success, struct)
#

###=====================================================================================================================
def Atom_Position_Mutation(i_generation, i_indiv, i_indiv_overall, struct0):
    n_try = 0
    n_shift = 1 + numpy.random.randint(low=0, high=shared.n_shift_max)
    while (True):
        rpos = numpy.copy(struct0.rpos)
        shift_list = []
        for i_shift in range(0, n_shift):
            i1_atom = numpy.random.randint(low=0, high=shared.n_atom)
            shift_list.append(i1_atom)

            rpos[i1_atom,:] = shared.margin +  numpy.random.random(3) * (1.0 - 2 * shared.margin)
        #
        if (hard_constraints.Check_Bond_Length(shared.atom_id, rpos)):
            (fp_key, fp_val) = fingerprint.RDF_Like_Fingerprint(shared.atom_id, rpos)
            if (fingerprint.Is_Unique(fp_val)):
                shared.fingerprint_archive.append(fp_val)
                success = True
                break
            #
        #
        n_try = n_try + 1
        if (n_try == shared.n_try_atom_position_mutation_max):
            success = False
            break
        #
    #
    shared.n_try_atom_position_mutation = shared.n_try_atom_position_mutation + n_try

    if (success):
        title = "%6i:   %4.4i.%4.4i" % (i_indiv_overall+1, i_generation+1, i_indiv+1)
        struct = structure.Structure(title=title)

        struct.rpos   = numpy.copy(rpos)
        struct.fp_key = numpy.copy(fp_key)
        struct.fp_val = numpy.copy(fp_val)

        struct.i_generation    = i_generation
        struct.i_indiv         = i_indiv
        struct.i_indiv_overall = i_indiv_overall

        history = "%4.4i.%4.4i   Atom_Position_Mutation   (%i)   %s" % (i_generation+1, i_indiv+1, struct0.i_indiv+1, str(shift_list))
        struct.history.append(history)
        struct.origin = "atom-position-mutation"
        struct.back = [struct0.i_indiv]

        struct.Set_Center()
    else:
        struct = None
    #

    return (success, struct)
#

###=====================================================================================================================


